#undef _MBCS
#define UNICODE
#define _UNICODE
#include <stdio.h>
#include <windows.h>
#include <tchar.h>

#include "ShiftSSA.h"
#include "resource.h"

extern HINSTANCE hInst;

BOOL GetSSAInfo( LPCWSTR lpwcPath, LPSSAINFO lpssa )
{
	WCHAR lpwcFileName[ MAX_PATH + 2 ];
	PathToFileName( lpwcPath, lpwcFileName, MAX_PATH + 1 );
	return GetSSAInfo( lpwcPath, lpwcFileName, lpssa );
}

BOOL GetSSAInfo( LPCWSTR lpwcPath, LPCWSTR lpwcFileTitle, LPSSAINFO lpssa )
{
	HWND hWnd = NULL;
	WCHAR msg[ LINE_BUFFER_SIZE * 2 ];
	UINT uCP = 0;
	ZeroMemory( lpssa, sizeof( SSAINFO ) );

	int iBom = DetectUTF16( lpwcPath );

	if( iBom == BOM_ERROR ) return FALSE;

	if( iBom ==  'L' ) wcscpy( lpssa->lpwcCharset, L"Unicode ( UTF-16 LE BOM )" );
	else if( iBom == 'l' ) wcscpy( lpssa->lpwcCharset, L"UTF-16 LE without BOM" );
	else if( iBom == 'B' ) wcscpy( lpssa->lpwcCharset, L"UTF-16 BE BOM" );
	else if( iBom == 'b' ) wcscpy( lpssa->lpwcCharset, L"UTF-16 BE without BOM" );
	else 
	{	
		uCP = DialogBoxParam( hInst, (LPCTSTR)IDD_SELECTCP, hWnd, (DLGPROC)SelectCP,
						( LPARAM ) lpwcFileTitle );
		if( ! uCP )
		{
			ZeroMemory( lpssa, sizeof( SSAINFO ) );
			return FALSE;
		}
		WCHAR lpwcCharset[ 1024 ] = L"";
		GetCPString( uCP, lpwcCharset );
		wcscpy( lpssa->lpwcCharset, lpwcCharset );
	}

	if( iBom == 'B' || iBom == 'b' ) 
	{
		ZeroMemory( lpssa, sizeof( SSAINFO ) );
		MessageBox( hWnd, L"For now, UTF-16 BE is not supported.", lpwcFileTitle, MB_ICONEXCLAMATION | MB_OK );
		return FALSE;
	}

	FILE * fin = _wfopen( lpwcPath, L"rb" );

	if( ! fin )
	{
		ZeroMemory( lpssa, sizeof( SSAINFO ) );
		MessageBox( hWnd, lpwcPath, L"!_wfopen", MB_ICONEXCLAMATION | MB_OK );
		return FALSE;
	}

	if( iBom == UTF16_LE_BOM || iBom == UTF16_BE_BOM )
	{
		fgetc( fin );
		fgetc( fin );
	}


	SSALINEW slw;
	WCHAR lpwcBuffer[ LINE_BUFFER_SIZE ] = L"";
	char lpMBStr[ LINE_BUFFER_SIZE ] = "";
	BOOL bVersion = FALSE;
	BOOL bSampleDialogue = FALSE;
	for( int i = 1;  i < 10000; i++ )
	{
		if( iBom == NOT_UTF16 )
		{
			if( ! fgets( lpMBStr, LINE_BUFFER_SIZE, fin ) ) break;
			if( lstrlenA( lpMBStr ) > LINE_BUFFER_SIZE - 2 )
			{
				ZeroMemory( lpssa, sizeof( SSAINFO ) );
				MessageBox( hWnd, L"Too long line, lstrlenA", L"", MB_OK | MB_ICONSTOP );
				fclose( fin );
				return FALSE;
			}
			if( ! MultiByteToWideChar(
				uCP, MB_ERR_INVALID_CHARS, lpMBStr, -1, lpwcBuffer, LINE_BUFFER_SIZE )
			)
			{
				ZeroMemory( lpssa, sizeof( SSAINFO ) );
				MessageBox( hWnd, L"! MultiByteToWideChar", L"", MB_OK | MB_ICONSTOP );
				fclose( fin );
				return FALSE;
			}
		}
		else 
		{
			if( ! fgetws( lpwcBuffer, LINE_BUFFER_SIZE, fin ) ) break;
		}
		
		if( wcslen( lpwcBuffer ) > LINE_BUFFER_SIZE - 2 )
		{
			ZeroMemory( lpssa, sizeof( SSAINFO ) );
			MessageBox( hWnd, L"Too long line.", APP_NAME, MB_OK | MB_ICONSTOP );
			fclose( fin );
			return FALSE;
		}

		int iResult = ReadSSALineW( lpwcBuffer, &slw );

		if( iResult == SSA_LINE_VERSION ) bVersion = GetSSAVersionFromHeader( lpwcBuffer, lpssa );
		else if( iResult == SSA_LINE_DIALOGUE )
		{
			if( ! bSampleDialogue ) bSampleDialogue = GetSSASampleDialogue( lpwcBuffer, lpssa );
			if( ! bVersion ) bVersion = GetSSAVersionFromDialogue( lpwcBuffer, lpssa );
			( lpssa->dwDialogues )++;
		}
		else if( iResult == SSA_LINE_EMBEDDED )
		{
			wsprintf( msg, L"Stopped at line %d:\r\n\r\n%s", i, lpwcBuffer );
			MessageBox( hWnd, msg, APP_NAME, MB_ICONEXCLAMATION | MB_OK );
			break;
		}
		else if( iResult == SSA_LINE_ERROR )
		{
			ZeroMemory( lpssa, sizeof( SSAINFO ) );
			wsprintf( msg, L"Abnormal data:\r\nLine %d\r\nFile: %s\r\n%s", i, lpwcFileTitle, lpwcBuffer );
			MessageBox( hWnd, msg, APP_NAME, MB_ICONSTOP | MB_OK );
			fclose( fin );
			return FALSE;
		}
	}

	fclose( fin );
	wcscpy( lpssa->lpwcFileName, lpwcFileTitle );
	wcscpy( lpssa->lpwcFilePath, lpwcPath );
	return TRUE;
}

BOOL GetSSASampleDialogue( LPWSTR lpwcLine, LPSSAINFO lpssa )
{
	SSALIST s;
	if( SplitW( lpwcLine, &s ) != 10 ) return FALSE;
	WCHAR str[ LINE_BUFFER_SIZE ];
	wcscpy( str, s.list[ 9 ] );
	ChompW( str );
	if( wcslen( str ) >= 30 ) 
	{
		str[ 30 ] = L'\0';
		wcscat( str, L"..." );
	}
	wcscpy( lpssa->lpwcSampleDialogue, str );
	return TRUE;
}
//int SplitW( LPCWSTR lpwcLine, LPWSTR[] list, DWORD dwBufferSize, int iListSize )


BOOL GetSSAVersionFromDialogue( LPCWSTR lpwcLine, LPSSAINFO lpssa )
{
	LPCWSTR str = L"Dialogue: Marked=";
	if( wcsncmp( lpwcLine, str, wcslen( str ) ) == 0 )
	{
		wcscpy( lpssa->lpwcVersion, L"SSA ( Sub Station Alpha )" );
	}
	else
	{
		wcscpy( lpssa->lpwcVersion, L"ASS ( Advanced Sub Station Alpha )" );
	}
	return TRUE;
}

BOOL GetSSAVersionFromHeader( LPCWSTR lpwcLine, LPSSAINFO lpssa )
{
	wcscpy( lpssa->lpwcVersion, L"" );
	WCHAR str[ LINE_BUFFER_SIZE ];
	if( wcslen( lpwcLine ) >= LINE_BUFFER_SIZE ) return FALSE;
	wcscpy( str, lpwcLine );
	ChompW( str );

	if( wcscmp( str, L"ScriptType: v4.00" ) == 0 )
	{
		wcscpy( lpssa->lpwcVersion, L"SSA ( Sub Station Alpha 4.00 )" );
		return TRUE;
	}
	else if( wcscmp( str, L"ScriptType: v4.00+" ) == 0 )
	{
		wcscpy( lpssa->lpwcVersion, L"ASS ( Advanced Sub Station Alpha 4.00+ )" );
		return TRUE;
	}
	else
	{
		wcscpy( lpssa->lpwcVersion, str );
		return TRUE;
	}
/*
	wcscpy( str, lpssa->lpwcFileName );
	CharLower( str );
	int len = wcslen( str );
	if( wcscmp( str[ len - 4 ], L".ssa" ) )
	{
		wcscpy( lpssa->lpwcVersion, L".ssa file" );
		return TRUE;
	}
	else if( wcscmp( str[ len - 4 ], L".ass" ) )
	{
		wcscpy( lpssa->lpwcVersion, L".ass file" );
		return TRUE;
	}
*/

}

VOID ChompW( LPWSTR str )
{
	int len = wcslen( str );
	if( str[ len - 2 ] == L'\r' ) str[ len - 2 ] = L'\0';
	else if( str[ len - 1 ] == L'\n' ) str[ len - 1 ] = L'\0';
}


int SplitW( LPWSTR lpwcLine, LPSSALIST ls )
{
	const int iBufferSize = LINE_BUFFER_SIZE;
	const int iListSize = 10;
	ZeroMemory( ls, sizeof( SSALIST) );
	LPWSTR lpwc[ 12 ];
	lpwc[ 0 ] = lpwcLine;
	const WCHAR wcSep = L',';
for( int i = 0; i < iListSize; i++ )
{
	if( i < iListSize - 1 )
	{
		lpwc[ i + 1 ] = wcschr( lpwc[ i ], wcSep );
		if( ! lpwc[ i + 1 ] )
		{
			if( wcslen( lpwc[ i ] ) < iBufferSize )
			{
				wcscpy( ls->list[ i ], lpwc[ i ] );
				return i + 1;
			}
			return i;
		}
		int len = wcslen( lpwc[ i ] ) - wcslen( lpwc[ i + 1 ] );
		if( len >= iBufferSize ) return i;
		wcsncpy( ls->list[ i ], lpwc[ i ], len );
		ls->list[ i ][ len ] = L'\0'; // for sure
		( lpwc[ i + 1 ] )++;
	}
	else
	{
		if( wcslen( lpwc[ i ] ) < iBufferSize )
		{
			wcscpy( ls->list[ i ], lpwc[ i ] );
			return i + 1;
		}
		return i;
	}
}
	return 0;
}

int ReadSSALineW( LPCWSTR lpwcLine, LPSSALINEW lpSSALineW )
{
	ZeroMemory( lpSSALineW, sizeof( SSALINEW ) );

//	WCHAR msg[ LINE_BUFFER_SIZE * 2 ];

	WCHAR lpwcTest[ 100 ] = L"";

	LPCWSTR lpwcF = L"[Fonts]";
	wcsncpy( lpwcTest, lpwcLine, wcslen( lpwcF ) );
	lpwcTest[ wcslen( lpwcF ) ] = L'\0';
	if( wcscmp( lpwcTest, lpwcF ) == 0 ) return SSA_LINE_EMBEDDED;

	LPCWSTR lpwcG = L"[Graphics]";
	wcsncpy( lpwcTest, lpwcLine, wcslen( lpwcG ) );
	lpwcTest[ wcslen( lpwcG ) ] = L'\0';
	if( wcscmp( lpwcTest, lpwcG ) == 0 ) return SSA_LINE_EMBEDDED;

	LPCWSTR lpwcVersionHeader = L"ScriptType:";
	wcsncpy( lpwcTest, lpwcLine, wcslen( lpwcVersionHeader ) );
	lpwcTest[ wcslen( lpwcVersionHeader ) ] = L'\0';
	if( wcscmp( lpwcTest, lpwcVersionHeader ) == 0 ) return SSA_LINE_VERSION;

	LPCWSTR lpwcDialogueHeader = L"Dialogue:";
	wcsncpy( lpwcTest, lpwcLine, wcslen( lpwcDialogueHeader ) );
	lpwcTest[ wcslen( lpwcDialogueHeader ) ] = L'\0';

	if( wcscmp( lpwcTest, lpwcDialogueHeader ) != 0 )
	{
		return SSA_LINE_SKIPPED;
	}

	WCHAR lpwcStartTime[ 128 ] = L"";
	WCHAR lpwcEndTime[ 128 ] = L"";

	WCHAR wcSep = L',';

	LPWSTR lpwc1 = wcschr( lpwcLine, wcSep );
	if( ! lpwc1 ) return SSA_LINE_ERROR;
	int len1 = wcslen( lpwcLine ) - wcslen( lpwc1 );
	if( len1 > 100 ) return SSA_LINE_ERROR;
	wcsncpy( lpSSALineW->lpwcHeader, lpwcLine, len1 );
	lpSSALineW->lpwcHeader[ len1 ] = L'\0';

	lpwc1++;

	LPWSTR lpwc2 = wcschr( lpwc1, wcSep );
	if( ! lpwc2 ) return SSA_LINE_ERROR;
	int len2 = wcslen( lpwc1 ) - wcslen( lpwc2 );
	if( len2 > 100 ) return SSA_LINE_ERROR;
	wcsncpy( lpwcStartTime, lpwc1, len2 );
	lpwcStartTime[ len2 ] = L'\0';
	double dStart = TimeStringToDoubleW( lpwcStartTime );
	if( dStart < 0.0 ) return SSA_LINE_ERROR;
	lpSSALineW->dStart = dStart;
	lpwc2++;

	LPWSTR lpwc3 = wcschr( lpwc2, wcSep );
	if( ! lpwc3 ) return SSA_LINE_ERROR;
	int len3 = wcslen( lpwc2 ) - wcslen( lpwc3 );
	if( len3 > 100 ) return SSA_LINE_ERROR;
	wcsncpy( lpwcEndTime, lpwc2, len3 );
	lpwcEndTime[ len3 ] = L'\0';
	double dEnd = TimeStringToDoubleW( lpwcEndTime );
	if( dEnd < 0.0 ) return SSA_LINE_ERROR;
	lpSSALineW->dEnd = dEnd;

	lpwc3++;

	int len4 = wcslen( lpwc3 );
	if( len1 > LINE_BUFFER_SIZE - 4 ) return SSA_LINE_ERROR;

	wcscpy( lpSSALineW->lpwcRemains, lpwc3 );
	lpSSALineW->lpwcRemains[ len4 ] = L'\0';

/*
	swprintf( msg, L"%s\r\n%s [%.2f]\r\n%s [%.2f]\r\n%s\r\n",
				lpwcHeader,
				lpwcStartTime, dStart,
				lpwcEndTime,   dEnd,
				lpwcRemains
	);
	MessageBoxW( NULL, msg, APP_NAME, MB_OK );
*/
	return SSA_LINE_DIALOGUE;
}


double TimeStringToDoubleW( LPCWSTR lpwcString )
{
	// 1:34:56.78

	WCHAR lpwcHours[ 20 ];
	WCHAR lpwcMinutes[ 20 ];
	WCHAR lpwcSeconds[ 20 ];
	WCHAR lpwcCentiseconds[ 20 ];
	double h, m, s, cs;

	int len = wcslen( lpwcString );

	if( len != 10 && len != 11 ) return -1.0;

	if( len == 10 )
	{
		if(
			! IsDigitW( lpwcString[ 0 ] )
			|| lpwcString[ 1 ] != L':'
			|| ! IsDigitW( lpwcString[ 2 ] )
			|| ! IsDigitW( lpwcString[ 3 ] )
			|| lpwcString[ 4 ] != L':'
			|| ! IsDigitW( lpwcString[ 5 ] )
			|| ! IsDigitW( lpwcString[ 6 ] )
			|| lpwcString[ 7 ] != L'.'
			|| ! IsDigitW( lpwcString[ 8 ] )
			|| ! IsDigitW( lpwcString[ 9 ] )
		) return -1.0;

		wcscpy( lpwcHours, &lpwcString[ 0 ] );
		lpwcHours[ 1 ] = L'\0';
		wcscpy( lpwcMinutes, &lpwcString[ 2 ] );
		lpwcMinutes[ 2 ] = L'\0';

		wcscpy( lpwcSeconds, &lpwcString[ 5 ] );
		lpwcSeconds[ 2 ] = L'\0';

		wcscpy( lpwcCentiseconds, &lpwcString[ 8 ] );
		lpwcCentiseconds[ 2 ] = L'\0';

	}
	else if( len == 11 )
	{
		if(
			! IsDigitW( lpwcString[ 0 ] )
			|| ! IsDigitW( lpwcString[ 1 ] )
			|| lpwcString[ 2 ] != L':'
			|| ! IsDigitW( lpwcString[ 3 ] )
			|| ! IsDigitW( lpwcString[ 4 ] )
			|| lpwcString[ 5 ] != L':'
			|| ! IsDigitW( lpwcString[ 6 ] )
			|| ! IsDigitW( lpwcString[ 7 ] )
			|| lpwcString[ 8 ] != L'.'
			|| ! IsDigitW( lpwcString[ 9 ] )
			|| ! IsDigitW( lpwcString[ 10 ] )
		) return -1.0;

		wcscpy( lpwcHours, &lpwcString[ 0 ] );
		lpwcHours[ 2 ] = L'\0';

		wcscpy( lpwcMinutes, &lpwcString[ 3 ] );
		lpwcMinutes[ 2 ] = L'\0';

		wcscpy( lpwcSeconds, &lpwcString[ 6 ] );
		lpwcSeconds[ 2 ] = L'\0';

		wcscpy( lpwcCentiseconds, &lpwcString[ 9 ] );
		lpwcCentiseconds[ 2 ] = L'\0';
	}
	else
	{
		return -1.0;
	}

	h = _wtoi( lpwcHours );
	m = _wtoi( lpwcMinutes );
	s = _wtoi( lpwcSeconds );
	cs = _wtoi( lpwcCentiseconds );

	if( m >= 60.0 || s >= 60.0 ) return -1.0;

	return h * 3600.0 + m * 60.0 + s + cs / 100.0;

}

BOOL IsDigitW( WCHAR wc )
{
	LPCWSTR lpwcStr = L"0123456789";
	return ( wcschr( lpwcStr, wc ) != NULL );
}
/*
double StringToDoubleW( LPWCSTR lpwcStr )
{
	if( wcslen( lpwcStr ) > 100 ) return 0.0;

}*/

int DetectUTF16( LPCTSTR lpszFileName )
{
	HANDLE hFile = CreateFile(
		lpszFileName,
		GENERIC_READ,
		0,
		NULL,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,
		NULL
	);

	if( hFile == INVALID_HANDLE_VALUE )
	{
		return BOM_ERROR;
	}

	BYTE lpBuffer[ 1024 ];
	ZeroMemory( lpBuffer, 1024 );
	DWORD dwBytes = 0;
	ReadFile( hFile, lpBuffer, 2, &dwBytes, NULL );
	CloseHandle( hFile );

	if( dwBytes != 2 )
	{
		return BOM_ERROR;
	}

	if( lpBuffer[ 0 ] == 0xFF && lpBuffer[ 1 ] == 0xFE )
	{
		return UTF16_LE_BOM;
	}
	else if( lpBuffer[ 0 ] == '[' && lpBuffer[ 1 ] == '\0' )
	{
		return UTF16_LE_NO_BOM;
	}
	else if( lpBuffer[ 0 ] == 0xFE && lpBuffer[ 1 ] == 0xFF )
	{
		return UTF16_BE_BOM;
	}
	else if( lpBuffer[ 0 ] == '\0' && lpBuffer[ 1 ] == '['  )
	{
		return UTF16_BE_NO_BOM;
	}

	return BOM_NOT_DETECTED;
}
