#undef _MBCS
#define UNICODE
#define _UNICODE
#include <stdio.h>
#include <windows.h>
#include <tchar.h>

#include <math.h>

#include "ShiftSSA.h"
#include "resource.h"

extern HINSTANCE hInst;

BOOL ShiftSSA( HWND hWnd, LPCWSTR lpwcFilePath, const LONG lShift, const double dFps,
			  const LONG lFrame, const BOOL bLimitGte )
{
	WCHAR msg[ LINE_BUFFER_SIZE * 2 ];
	if( dFps <= 0.0 )
	{
		MessageBox( hWnd, L"Invalid parameter(s).", APP_NAME, MB_ICONEXCLAMATION | MB_OK );
		return FALSE;
	}
	const double dShift = lShift / dFps;
	double dStart, dEnd;
	const double dTimeLimit = lFrame / dFps;
	if( fabs( dShift ) > 100 * 3600.0
		|| dFps > 120.0 || wcslen( lpwcFilePath ) == 0 
		|| lFrame < 0
	)
	{
		MessageBox( hWnd, L"Invalid parameter(s).", APP_NAME, MB_ICONEXCLAMATION | MB_OK );
		return FALSE;
	}
	double sec = fabs( dShift );
	int h = (int) floor( sec / 3600.0 );
	sec -= 3600 * h;
	int m = (int) floor( sec / 60.0 );
	sec -= 60 * m;
	double sec2 = dTimeLimit;
	int h2 = (int) floor( sec2 / 3600.0 );
	sec2 -= 3600 * h2;
	int m2 = (int) floor( sec2 / 60.0 );
	sec2 -= 60 * m2;
	swprintf( msg, L"%s\r\n\r\nby %ld frame(s) at %f fps (approx. %d:%02d:%05.2f)\r\n\r\n"
					L"if the original Start Frame # %s %ld (approx. %d:%02d:%05.2f)",
						lpwcFilePath,
						lShift, dFps, h, m, sec,
						( bLimitGte ? L">=" : L"<=" ), lFrame, h2, m2, sec2	
					);
	if( IDCANCEL == MessageBox( hWnd, msg, L"Shifting", MB_ICONQUESTION | MB_OKCANCEL ) )
	{
		return FALSE;
	}

	int iBom = DetectUTF16( lpwcFilePath );

	FILE * fin = _wfopen( lpwcFilePath, L"rb" );
	if( ! fin )
	{
		MessageBox( hWnd, lpwcFilePath, L"!_wfopen", MB_ICONEXCLAMATION | MB_OK );
		return FALSE;
	}

	WCHAR lpwcTmpPath[ MAX_PATH + 1 ];
	WCHAR lpwcTmpFileName[ MAX_PATH + 1 ];
	WCHAR lpwcBakFileName[ MAX_PATH + 1 ];

	GetTempPath( MAX_PATH, lpwcTmpPath );
	GetTempFileName( lpwcTmpPath, L"S$4", 0, lpwcTmpFileName );

	GetBakFileName( lpwcFilePath, lpwcBakFileName, MAX_PATH );

	FILE * fout = _wfopen( lpwcTmpFileName, L"wb" );
	if( !fout )
	{
		fclose( fin );
		MessageBox( hWnd, lpwcTmpPath, L"!_wfopen", MB_ICONEXCLAMATION | MB_OK );
		return FALSE;
	}

	DWORD dwCounter = 0;
	DWORD dwCounterShifted = 0;
	DWORD dwCounterNotShifted = 0;

	SSALINEW slw;
	WCHAR lpBuffer[ LINE_BUFFER_SIZE ] = L"";
	char lpMBStr[ LINE_BUFFER_SIZE ] = "";
	UINT uCP = 0;

	if( iBom == UTF16_LE_BOM || iBom == UTF16_LE_NO_BOM )
	{
		if( iBom == UTF16_LE_BOM )
		{
			char c = fgetc( fin );
			fputc( c, fout );
			c = fgetc( fin );
			fputc( c, fout );
		}
	}
	else if( iBom == UTF16_BE_BOM || iBom == UTF16_BE_NO_BOM )
	{
		fclose( fin );
		fclose( fout );
		DeleteFile( lpwcTmpFileName );
		MessageBox( hWnd, L"UTF-16 BE not supported yet. Use UTF-16 LE.", APP_NAME, MB_OK );
		return FALSE;
	}
	else
	{
		WCHAR filename[ MAX_PATH ];
		PathToFileName( lpwcFilePath, filename, MAX_PATH );
		uCP = DialogBoxParam( hInst, (LPCTSTR)IDD_SELECTCP, hWnd, (DLGPROC)SelectCP,
						( LPARAM ) filename );
		if( uCP == 0 )
		{
			fclose( fin );
			fclose( fout );
			DeleteFile( lpwcTmpFileName );
			return FALSE;
		}
	}

	for( int i = 1; i < 10000; i++ )
	{
		if( iBom == NOT_UTF16 )
		{
			if( ! fgets( lpMBStr, LINE_BUFFER_SIZE, fin ) ) break;
			if( lstrlenA( lpMBStr ) > LINE_BUFFER_SIZE - 2 )
			{
				MessageBox( hWnd, L"Too long line, lstrlenA", L"", MB_OK | MB_ICONSTOP );
				fclose( fin );
				fclose( fout );
				DeleteFile( lpwcTmpFileName );
				return FALSE;
			}
			if( ! MultiByteToWideChar(
				uCP, MB_ERR_INVALID_CHARS, lpMBStr, -1, lpBuffer, LINE_BUFFER_SIZE )
			)
			{
				MessageBox( hWnd, L"! MultiByteToWideChar", L"", MB_OK | MB_ICONSTOP );
				fclose( fin );
				fclose( fout );
				DeleteFile( lpwcTmpFileName );
				return FALSE;
			}
		}
		else
		{
			if( ! fgetws( lpBuffer, LINE_BUFFER_SIZE, fin ) ) break;
		}

		if( wcslen( lpBuffer ) > LINE_BUFFER_SIZE - 2 )
			{
				fclose( fin );
				fclose( fout );
				DeleteFile( lpwcTmpFileName );
				MessageBox( hWnd, L"Too long line.", APP_NAME, MB_OK | MB_ICONSTOP );
				return FALSE;
			}
			int iResult = ReadSSALineW( lpBuffer, &slw );

			if( iResult == SSA_LINE_SKIPPED || iResult == SSA_LINE_VERSION )
			{
				if( PutLineCP( lpBuffer, fout, uCP ) ) continue;
				fclose( fin );
				fclose( fout );
				DeleteFile( lpwcTmpFileName );
				MessageBox( hWnd, L"! PutLineCP while skipping.", APP_NAME, MB_OK | MB_ICONSTOP );
				return FALSE;
			}
			else if( iResult == SSA_LINE_DIALOGUE )
			{
				WCHAR lpwcStartNormalized[ 128 ];
				WCHAR lpwcEndNormalized[ 128 ];
				dStart = slw.dStart;
				dEnd = slw.dEnd;
				if( bLimitGte && dStart >= dTimeLimit || ! bLimitGte && dStart <= dTimeLimit )
				{
					dwCounterShifted++;
					dStart += dShift;
					dEnd   += dShift;
				}
				else
				{
					dwCounterNotShifted++;
				}

				if( dStart < 0.0 || dEnd < 0.0 ||
					dStart > 100 * 3600.0 || dEnd > 100 * 3600.0
				)
				{
					fclose( fin );
					fclose( fout );
					DeleteFile( lpwcTmpFileName );
					swprintf( msg, L"The specified shift is too large in absolute value for:\r\n"
						L"Line %d\r\n%s",
						i, lpBuffer );

					MessageBox( hWnd, msg, L"Error", MB_ICONSTOP | MB_OK );
					return FALSE;
				}
				if(
					! NormalizeSSAW( dStart, dFps, lpwcStartNormalized, 128 )
					||
					! NormalizeSSAW( dEnd, dFps, lpwcEndNormalized, 128 )
				)
				{
					fclose( fin );
					fclose( fout );
					DeleteFile( lpwcTmpFileName );
					MessageBox( hWnd, lpBuffer, L"Error", MB_ICONEXCLAMATION | MB_OK );
					return FALSE;
				}

				WCHAR wcstr[ LINE_BUFFER_SIZE * 2 ];
				swprintf( wcstr,
					L"%s,%s,%s,%s",
							slw.lpwcHeader,
							lpwcStartNormalized,
							lpwcEndNormalized,
							slw.lpwcRemains
				);
				if( ! PutLineCP( wcstr, fout, uCP ) )
				{
					fclose( fin );
					fclose( fout );
					DeleteFile( lpwcTmpFileName );
					MessageBox( hWnd, lpBuffer, L"! PutLineCP", MB_ICONEXCLAMATION | MB_OK );
					return FALSE;
				}
				dwCounter++;
			}
			else
			{
				MessageBox( hWnd, lpBuffer, L"Stopped", MB_ICONINFORMATION | MB_OK );
				break;
			}
	}

	fclose( fin );
	fclose( fout );

	if( ! CopyFile( lpwcFilePath, lpwcBakFileName, TRUE ) )
	{
		wsprintf( msg,
			L"Failed to back up.\r\nThe original file is not changed.\r\nThe resulted file:\r\n%s",
			lpwcTmpFileName );
		MessageBox( hWnd, msg, APP_NAME, MB_OK );
		return TRUE;
	}

	DeleteFile( lpwcFilePath );
	if( ! CopyFile( lpwcTmpFileName, lpwcFilePath, FALSE ) )
	{
		wsprintf( msg,
			L"Failed to copy.\r\nThe original file is backed up as:\r\n%s\r\nThe resulted file:\r\n%s",
			lpwcBakFileName, lpwcTmpFileName );
		MessageBox( hWnd, msg, APP_NAME, MB_OK );
		return TRUE;
	}
	DeleteFile( lpwcTmpFileName );

	wsprintf( msg,
		L"%lu Dialogue(s) processed:\r\n\r\n"
		L"%lu Dialogue(s) shifted, %lu Dialogue(s) not shifted.\r\n\r\n"
		L"The original file backed up as:\r\n\r\n%s",
		dwCounter,
		dwCounterShifted,
		dwCounterNotShifted,
		lpwcBakFileName );
	MessageBox( hWnd, msg, APP_NAME, MB_ICONINFORMATION | MB_OK );
	return TRUE;
}


double NormalizeSSA( double dSeconds, double dFps )
{
	if( dSeconds <= 0.0 ) return 0.0;
	DWORD dwFrames = (DWORD) ceil( dSeconds * dFps );
	double t = dwFrames / dFps;
	DWORD dwSeconds = (DWORD) floor( t );
	DWORD dwMilliseconds = (DWORD) floor( ( t - dwSeconds ) * 1000.0 );
	DWORD r = dwMilliseconds % 10;
	if( r == 0 ) r = 10;

	if( r > dwMilliseconds )
	{
		dwSeconds--;
		dwMilliseconds += 1000;
	}

//	WCHAR str[1000];
//	swprintf( str, L"f# %lu\r\ns %lu\r\nms %lu\r\nr %lu\r\n", dwFrames, dwSeconds, dwMilliseconds, r );
//	MessageBox( NULL, str, APP_NAME, MB_OK );
	return dwSeconds + ( dwMilliseconds - r ) / 1000.0;
}

BOOL NormalizeSSAW( double dSeconds, double dFPS, LPWSTR lpwcString, DWORD dwBufferSize )
{
	if( dSeconds < 0.0 || dSeconds >= 100 * 3600.0 || dwBufferSize < 16 ) 
	{
		wcscpy( lpwcString, L"" );
		return FALSE;
	}

	double t = NormalizeSSA( dSeconds, dFPS );
	int h = ( int ) floor( t / 3600.0 );
if( h > 99 )
{
	return FALSE;
}
	t -= h * 3600.0;
	int m = ( int ) floor( t / 60.0 );
	t -= m * 60;
	swprintf( lpwcString, L"%d:%02d:%05.2f", h , m , t );

	return TRUE;
}

LRESULT CALLBACK About(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	static HBRUSH hBrush = NULL;
	switch (message)
	{
		case WM_CTLCOLORBTN:
		case WM_CTLCOLORDLG:
		case WM_CTLCOLORSTATIC:
		{
			if( ! hBrush )
			{
				hBrush = CreateSolidBrush( RGB( 0xEE, 0xEE, 0xDD ) );
			}
//			SetBkMode( (HDC) wParam, TRANSPARENT );
			SetBkColor( (HDC) wParam, RGB( 0xEE, 0xEE, 0xDD ) );
			return ( LRESULT ) hBrush;
			break;
		}
		case WM_INITDIALOG:
		{
			WCHAR msg[ 1024 * 4 ];
			wsprintf( msg, L"Compiled at: %hs, %hs +0900", __TIME__, __DATE__ );
			SetDlgItemText( hDlg, IDC_VERSION , APP_NAME );
			SetDlgItemText( hDlg, IDC_DATETIME , msg );
			wsprintf( msg, 
				L"%s: Frame-accurate time shift for SSA/ASS\r\n"
				L"Copyright (C) 2004, %s\r\n\r\n", APP_NAME, MY_URL );
			wcscat( msg, TEXT_GPL );
			SetDlgItemText( hDlg, IDC_EDIT1 , msg );
			return TRUE;
		}
		case WM_COMMAND:
		if ( LOWORD( wParam ) == IDOK || LOWORD( wParam ) == IDCANCEL ) 
		{
			if( hBrush )
			{
				DeleteObject( hBrush );
				hBrush = NULL;
			}
			EndDialog( hDlg, 0 );
		}
			return TRUE;
	}
    return FALSE;
}

